#ifndef MUBLAS_H
#define MUBLAS_H

#include <cuComplex.h>

// ***** FORMAT *****

struct __align__(8) cudfreal
{
	float x;
	float y;
};

struct __align__(16) cuddreal
{
	double x;
	double y;
};

struct __align__(16) cudfcomplex
{
	cudfreal x;
	cudfreal y;
};

struct cuddcomplex
{
	cuddreal x;
	cuddreal y;
};

// cudfreal ----------------------------------
__host__ __device__ __forceinline__ cudfreal
make_cudfreal 
	(const float h, const float l)
{
	cudfreal v;
	v.x = h;
	v.y = l;
	return v;
}
// cudfreal ----------------------------------

// cudfcomplex ----------------------------------
__host__ __device__ __forceinline__ cudfcomplex
make_cudfcomplex 
	(const cudfreal re, const cudfreal im)
{
	cudfcomplex v;
	v.x = re;
	v.y = im;
	return v;
}

__host__ __device__ __forceinline__ cudfcomplex
make_cudfcomplex
	(const float reh, const float rel, const float imh, const float iml)
{
	cudfcomplex v;
	v.x.x = reh;
	v.x.y = rel;
	v.y.x = imh;
	v.y.y = iml;
	return v;
}
// cudfcomplex ----------------------------------

// cuddreal ----------------------------------
__host__ __device__ __forceinline__ cuddreal
make_cuddreal 
	(const double h, const double l)
{
	cuddreal v;
	v.x = h;
	v.y = l;
	return v;
}

__host__ __device__ __forceinline__ cuddreal
make_cuddreal 
	(const double h)
{
	cuddreal v;
	v.x = h;
	v.y = 0.;
	return v;
}
// cuddreal ----------------------------------

// cuddcomplex ----------------------------------
__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex 
	(const cuddreal re, const cuddreal im)
{
	cuddcomplex v;
	v.x = re;
	v.y = im;
	return v;
}

__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex 
	(const double re, const double im)
{
	cuddcomplex v;
	v.x.x = re;
	v.x.y = 0.;
	v.y.x = im;
	v.y.y = 0.;
	return v;
}

__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex
	(const double reh, const double rel, const double imh, const double iml)
{
	cuddcomplex v;
	v.x.x = reh;
	v.x.y = rel;
	v.y.x = imh;
	v.y.y = iml;
	return v;
}
// cuddcomplex ----------------------------------

#define NGPU_MAX 8

struct mublasMgpuVectorHandle_t {
	size_t n[NGPU_MAX];
	size_t word_size;
	uint32_t ngpus;
	void* devAddr[NGPU_MAX];
};

struct mublasMgpuMatrixHandle_t {
	size_t m[NGPU_MAX];
	size_t n[NGPU_MAX];
	size_t ld[NGPU_MAX];
	size_t word_size;
	uint32_t ngpus;
	void* devAddr[NGPU_MAX];
};

struct mublasHandle_t {
	cudaStream_t streamId;
	cudaDeviceProp devprop;

	// mublas parameter
	int32_t mublasVersion;	// 00.00.00
	int32_t mublaserror;	// error handler
	int32_t count;			// 
	float *work_addr;
	float *work_addr2;
	int32_t debug;			// 
	int32_t autotuning;		// 

	// device parameter
	int32_t driverVersion; 
	int32_t runtimeVersion;
	int32_t cc;

	// architecture parameter
	size_t WrpPerMpPmax;		// Physical Max of Warps per Multiprocessor
	size_t BlkPerMpPmax;		// Physical Max of Blocks per Multiprocessor
	size_t RegUnit;			// Register Allocation Unit
	size_t ShmUnit;			// Shared Memory Allocation Unit
	size_t WrpGran;			// Warp Allocation Granularity
	size_t Cores;			// # of CUDA Cores per Multiprocessor
	size_t DPCores;		// # of DP Cores per Multiprocessor

	// kernel parameters
	int32_t kernelnumber;
	int32_t kerneldebug;
	cudaFuncAttributes funcatb;
	void (*func_getShmPerBlk)(mublasHandle_t* mh);
	void (*func_getBlkPerGrd)(mublasHandle_t* mh);
	enum cudaFuncCache func_cache;
	size_t N_T;
	size_t BlkDim;
	float WrpOcp;	// Warp Occupancy
	float BlkOcp;	// Block Occupancy
	float GrdOcp;	// Grid Occupancy
	size_t Bx;
	size_t By;
	size_t Tx;
	size_t Ty;
	size_t m;
	size_t n;
	size_t k;
	size_t ShmPerBlk;
	size_t ShmPerBlkMax;
	size_t BlkPerMpMax;
	size_t N_one;

	// tuning parameter
	float WrpOcpSuf;
	float BlkOcpSuf;
	float GrdOcpSuf;
	size_t TxMin;
	size_t TxMax;
	size_t TxStatic;
	size_t TyStatic;
	size_t ThMin;
	size_t ThMax;
	int32_t TxStep;
	int32_t ThStep;
	float TyPerTxMax;
	int32_t TxModTyZero;
};

// mublas function
void mublasCreate (mublasHandle_t*);
void mublasCreateSimple (mublasHandle_t*);
void mublasSetStream (mublasHandle_t*, cudaStream_t);
void mublasDestroy (mublasHandle_t*);
void mublasDestroySimple (mublasHandle_t*);

// Multi GPU
void mublasSetMatrixMgpu1D (size_t, size_t, void*, size_t, void**, size_t, uint32_t, size_t, size_t);
void mublasSetVectorMgpu1D (size_t, void*, void**, uint32_t, size_t, size_t);
void mublasGetVectorMgpu1D (size_t, void**, void*, uint32_t, size_t, size_t);
void mublasMallocMatrixMgpu1D (size_t, size_t, void**, size_t, uint32_t, size_t, size_t);

// Multi GPU (old)
/*
void mublasMallocVectorMgpu (mublasMgpuVectorHandle_t*, size_t, size_t, size_t);
void mublasMallocVectorMgpuDistributed (mublasMgpuVectorHandle_t*, size_t, size_t, size_t);
void mublasMallocMatrixMgpuDistributed (mublasMgpuMatrixHandle_t*, char, size_t, size_t, size_t, size_t);
void mublasSetVectorMgpuDistributed (mublasMgpuVectorHandle_t*, void*);
void mublasSetVectorMgpu (mublasMgpuVectorHandle_t*, void*);
void mublasSetMatrixMgpuDistributed (mublasMgpuMatrixHandle_t*, char, void*, size_t);
void mublasGetVectorMgpuDistributed (mublasMgpuVectorHandle_t*, void*);
void mublasFreeMatrixMgpu (mublasMgpuMatrixHandle_t*);
void mublasFreeVectorMgpu (mublasMgpuVectorHandle_t*);
*/

// GEAM
extern int32_t mublasSgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const float*, const float*, const size_t, const float*, const float*, const size_t, float*, const size_t);
extern int32_t mublasDgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const double*, const double*, const size_t, const double*, const double*, const size_t, double*, const size_t);
extern int32_t mublasCgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const cuFloatComplex*, const cuFloatComplex*, const size_t, const cuFloatComplex*, const cuFloatComplex*, const size_t, cuFloatComplex*, const size_t);
extern int32_t mublasZgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const size_t, cuDoubleComplex*, const size_t);
extern int32_t mublasWgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const cuddreal*, const cuddreal*, const size_t, const cuddreal*, const cuddreal*, const size_t, cuddreal*, const size_t);
extern int32_t mublasUgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const cuddcomplex*, const cuddcomplex*, const size_t, const cuddcomplex*, const cuddcomplex*, const size_t, cuddcomplex*, const size_t);
extern int32_t mublasBgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const cudfreal*, const cudfreal*, const size_t, const cudfreal*, const cudfreal*, const size_t, cudfreal*, const size_t);
extern int32_t mublasVgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const cudfcomplex*, const cudfcomplex*, const size_t, const cudfcomplex*, const cudfcomplex*, const size_t, cudfcomplex*, const size_t);
extern int32_t mublasDBgeam (mublasHandle_t*, const char, const char, const size_t, const size_t, const double*, const double*, const size_t, const double*, const double*, const size_t, double*, const size_t);

// GEMM
extern int32_t mublasSgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const float*, const float*, const size_t, const float*, const size_t, const float*, float*, const size_t);
extern int32_t mublasDgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);
extern int32_t mublasCgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const cuFloatComplex*, const cuFloatComplex*, const size_t, const cuFloatComplex*, const size_t, const cuFloatComplex*, cuFloatComplex*, const size_t);
extern int32_t mublasZgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, cuDoubleComplex*, const size_t);
extern int32_t mublasWgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const cuddreal*, const cuddreal*, const size_t, const cuddreal*, const size_t, const cuddreal*, cuddreal*, const size_t);
extern int32_t mublasUgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const cuddcomplex*, const cuddcomplex*, const size_t, const cuddcomplex*, const size_t, const cuddcomplex*, cuddcomplex*, const size_t);
extern int32_t mublasBgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const cudfreal*, const cudfreal*, const size_t, const cudfreal*, const size_t, const cudfreal*, cudfreal*, const size_t);
extern int32_t mublasVgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const cudfcomplex*, const cudfcomplex*, const size_t, const cudfcomplex*, const size_t, const cudfcomplex*, cudfcomplex*, const size_t);
extern int32_t mublasDBgemm (mublasHandle_t*, const char, const char, const size_t, const size_t, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);

// GEMV
extern int32_t mublasSgemv (mublasHandle_t*, const char, const size_t, const size_t, const float*, const float*, const size_t, const float*, const size_t, const float*, float*, const size_t);
extern int32_t mublasDgemv (mublasHandle_t*, const char, const size_t, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);
extern int32_t mublasCgemv (mublasHandle_t*, const char, const size_t, const size_t, const cuFloatComplex*, const cuFloatComplex*, const size_t, const cuFloatComplex*, const size_t, const cuFloatComplex*, cuFloatComplex*, const size_t);
extern int32_t mublasZgemv (mublasHandle_t*, const char, const size_t, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, cuDoubleComplex*, const size_t);
extern int32_t mublasWgemv (mublasHandle_t*, const char, const size_t, const size_t, const cuddreal*, const cuddreal*, const size_t, const cuddreal*, const size_t, const cuddreal*, cuddreal*, const size_t);
extern int32_t mublasUgemv (mublasHandle_t*, const char, const size_t, const size_t, const cuddcomplex*, const cuddcomplex*, const size_t, const cuddcomplex*, const size_t, const cuddcomplex*, cuddcomplex*, const size_t);
extern int32_t mublasBgemv (mublasHandle_t*, const char, const size_t, const size_t, const cudfreal*, const cudfreal*, const size_t, const cudfreal*, const size_t, const cudfreal*, cudfreal*, const size_t);
extern int32_t mublasVgemv (mublasHandle_t*, const char, const size_t, const size_t, const cudfcomplex*, const cudfcomplex*, const size_t, const cudfcomplex*, const size_t, const cudfcomplex*, cudfcomplex*, const size_t);
extern int32_t mublasDBgemv (mublasHandle_t*, const char, const size_t, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);

extern int32_t mublasSgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const float, float**, const size_t, float**, const size_t, const float, float**, const size_t, const uint32_t, const size_t);
extern int32_t mublasDgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const double, double**, const size_t, double**, const size_t, const double, double**, const size_t, const uint32_t, const size_t);
extern int32_t mublasCgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const cuFloatComplex, cuFloatComplex**, const size_t, cuFloatComplex**, const size_t, const cuFloatComplex, cuFloatComplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasZgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const cuDoubleComplex, cuDoubleComplex**, const size_t, cuDoubleComplex**, const size_t, const cuDoubleComplex, cuDoubleComplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasWgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const cuddreal, cuddreal**, const size_t, cuddreal**, const size_t, const cuddreal, cuddreal**, const size_t, const uint32_t, const size_t);
extern int32_t mublasUgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const cuddcomplex, cuddcomplex**, const size_t, cuddcomplex**, const size_t, const cuddcomplex, cuddcomplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasBgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const cudfreal, cudfreal**, const size_t, cudfreal**, const size_t, const cudfreal, cudfreal**, const size_t, const uint32_t, const size_t);
extern int32_t mublasVgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const cudfcomplex, cudfcomplex**, const size_t, cudfcomplex**, const size_t, const cudfcomplex, cudfcomplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasDBgemvMgpuAsync (mublasHandle_t*, const char, const size_t, const size_t, const double, double**, const size_t, double**, const size_t, const double, double**, const size_t, const uint32_t, const size_t);

extern int32_t mublasSgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const float*, float**, const size_t, float**, const size_t, const float*, float**, const size_t, const uint32_t, const size_t);
extern int32_t mublasDgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const double*, double**, const size_t, double**, const size_t, const double*, double**, const size_t, const uint32_t, const size_t);
extern int32_t mublasCgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const cuFloatComplex*, cuFloatComplex**, const size_t, cuFloatComplex**, const size_t, const cuFloatComplex*, cuFloatComplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasZgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const cuDoubleComplex*, cuDoubleComplex**, const size_t, cuDoubleComplex**, const size_t, const cuDoubleComplex*, cuDoubleComplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasWgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const cuddreal*, cuddreal**, const size_t, cuddreal**, const size_t, const cuddreal*, cuddreal**, const size_t, const uint32_t, const size_t);
extern int32_t mublasUgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const cuddcomplex*, cuddcomplex**, const size_t, cuddcomplex**, const size_t, const cuddcomplex*, cuddcomplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasBgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const cudfreal*, cudfreal**, const size_t, cudfreal**, const size_t, const cudfreal*, cudfreal**, const size_t, const uint32_t, const size_t);
extern int32_t mublasVgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const cudfcomplex*, cudfcomplex**, const size_t, cudfcomplex**, const size_t, const cudfcomplex*, cudfcomplex**, const size_t, const uint32_t, const size_t);
extern int32_t mublasDBgemvMgpu (mublasHandle_t*, const char, const size_t, const size_t, const double*, double**, const size_t, double**, const size_t, const double*, double**, const size_t, const uint32_t, const size_t);

// SYMV
extern int32_t mublasSsymv (mublasHandle_t*, const char, const size_t, const float*, const float*, const size_t, const float*, const size_t, const float*, float*, const size_t);
extern int32_t mublasDsymv (mublasHandle_t*, const char, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);
extern int32_t mublasWsymv (mublasHandle_t*, const char, const size_t, const cuddreal*, const cuddreal*, const size_t, const cuddreal*, const size_t, const cuddreal*, cuddreal*, const size_t);
extern int32_t mublasBsymv (mublasHandle_t*, const char, const size_t, const cudfreal*, const cudfreal*, const size_t, const cudfreal*, const size_t, const cudfreal*, cudfreal*, const size_t);
extern int32_t mublasDBsymv (mublasHandle_t*, const char, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);

// HEMV
extern int32_t mublasChemv (mublasHandle_t*, const char, const size_t, const cuFloatComplex*, const cuFloatComplex*, const size_t, const cuFloatComplex*, const size_t, const cuFloatComplex*, cuFloatComplex*, const size_t);
extern int32_t mublasZhemv (mublasHandle_t*, const char, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, cuDoubleComplex*, const size_t);
extern int32_t mublasUhemv (mublasHandle_t*, const char, const size_t, const cuddcomplex*, const cuddcomplex*, const size_t, const cuddcomplex*, const size_t, const cuddcomplex*, cuddcomplex*, const size_t);
extern int32_t mublasVhemv (mublasHandle_t*, const char, const size_t, const cudfcomplex*, const cudfcomplex*, const size_t, const cudfcomplex*, const size_t, const cudfcomplex*, cudfcomplex*, const size_t);

// TRMV
extern int32_t mublasStrmv (mublasHandle_t*, const char, const char, const char, const size_t, const float*, const size_t, float*, const size_t);
extern int32_t mublasDtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const double*, const size_t, double*, const size_t);
extern int32_t mublasCtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const cuFloatComplex*, const size_t, cuFloatComplex*, const size_t);
extern int32_t mublasZtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const cuDoubleComplex*, const size_t, cuDoubleComplex*, const size_t);
extern int32_t mublasWtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const cuddreal*, const size_t, cuddreal*, const size_t);
extern int32_t mublasUtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const cuddcomplex*, const size_t, cuddcomplex*, const size_t);
extern int32_t mublasBtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const cudfreal*, const size_t, cudfreal*, const size_t);
extern int32_t mublasVtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const cudfcomplex*, const size_t, cudfcomplex*, const size_t);
extern int32_t mublasDBtrmv (mublasHandle_t*, const char, const char, const char, const size_t, const double*, const size_t, double*, const size_t);

// AXPY
extern int32_t mublasSaxpy (mublasHandle_t*, const size_t, const float*, const float*, const size_t, float*, const size_t);
extern int32_t mublasDaxpy (mublasHandle_t*, const size_t, const double*, const double*, const size_t, double*, const size_t);
extern int32_t mublasCaxpy (mublasHandle_t*, const size_t, const cuFloatComplex*, const cuFloatComplex*, const size_t, cuFloatComplex*, const size_t);
extern int32_t mublasZaxpy (mublasHandle_t*, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const size_t, cuDoubleComplex*, const size_t);
extern int32_t mublasWaxpy (mublasHandle_t*, const size_t, const cuddreal*, const cuddreal*, const size_t, cuddreal*, const size_t);
extern int32_t mublasUaxpy (mublasHandle_t*, const size_t, const cuddcomplex*, const cuddcomplex*, const size_t, cuddcomplex*, const size_t);
extern int32_t mublasBaxpy (mublasHandle_t*, const size_t, const cudfreal*, const cudfreal*, const size_t, cudfreal*, const size_t);
extern int32_t mublasVaxpy (mublasHandle_t*, const size_t, const cudfcomplex*, const cudfcomplex*, const size_t, cudfcomplex*, const size_t);
extern int32_t mublasDBaxpy (mublasHandle_t*, const size_t, const double*, const double*, const size_t, double*, const size_t);

// DOT
extern int32_t mublasSdot (mublasHandle_t*, const size_t, const float*, const size_t, const float*, const size_t, float*);
extern int32_t mublasDdot (mublasHandle_t*, const size_t, const double*, const size_t, const double*, const size_t, double*);
extern int32_t mublasCdot (mublasHandle_t*, const size_t, const cuFloatComplex*, const size_t, const cuFloatComplex*, const size_t, cuFloatComplex*);
extern int32_t mublasZdot (mublasHandle_t*, const size_t, const cuDoubleComplex*, const size_t, const cuDoubleComplex*, const size_t, cuDoubleComplex*);
extern int32_t mublasWdot (mublasHandle_t*, const size_t, const cuddreal*, const size_t, const cuddreal*, const size_t, cuddreal*);
extern int32_t mublasUdot (mublasHandle_t*, const size_t, const cuddcomplex*, const size_t, const cuddcomplex*, const size_t, cuddcomplex*);
extern int32_t mublasBdot (mublasHandle_t*, const size_t, const cudfreal*, const size_t, const cudfreal*, const size_t, cudfreal*);
extern int32_t mublasVdot (mublasHandle_t*, const size_t, const cudfcomplex*, const size_t, const cudfcomplex*, const size_t, cudfcomplex*);
extern int32_t mublasDBdot (mublasHandle_t*, const size_t, const double*, const size_t, const double*, const size_t, double*);

// COPY
extern int32_t mublasScopy (mublasHandle_t*, const size_t, const float*, const size_t, float*, const size_t);
extern int32_t mublasDcopy (mublasHandle_t*, const size_t, const double*, const size_t, double*, const size_t);
extern int32_t mublasCcopy (mublasHandle_t*, const size_t, const cuFloatComplex*, const size_t, cuFloatComplex*, const size_t);
extern int32_t mublasZcopy (mublasHandle_t*, const size_t, const cuDoubleComplex*, const size_t, cuDoubleComplex*, const size_t);
extern int32_t mublasWcopy (mublasHandle_t*, const size_t, const cuddreal*, const size_t, cuddreal*, const size_t);
extern int32_t mublasUcopy (mublasHandle_t*, const size_t, const cuddcomplex*, const size_t, cuddcomplex*, const size_t);
extern int32_t mublasBcopy (mublasHandle_t*, const size_t, const cudfreal*, const size_t, cudfreal*, const size_t);
extern int32_t mublasVcopy (mublasHandle_t*, const size_t, const cudfcomplex*, const size_t, cudfcomplex*, const size_t);
extern int32_t mublasDBcopy (mublasHandle_t*, const size_t, const double*, const size_t, double*, const size_t);

// SCAL
extern int32_t mublasSscal (mublasHandle_t*, const size_t, const float*, float*, const size_t);
extern int32_t mublasDscal (mublasHandle_t*, const size_t, const double*, double*, const size_t);
extern int32_t mublasCscal (mublasHandle_t*, const size_t, const cuFloatComplex*, cuFloatComplex*, const size_t);
extern int32_t mublasZscal (mublasHandle_t*, const size_t, const cuDoubleComplex*, cuDoubleComplex*, const size_t);
extern int32_t mublasWscal (mublasHandle_t*, const size_t, const cuddreal*, cuddreal*, const size_t);
extern int32_t mublasUscal (mublasHandle_t*, const size_t, const cuddcomplex*, cuddcomplex*, const size_t);
extern int32_t mublasBscal (mublasHandle_t*, const size_t, const cudfreal*, cudfreal*, const size_t);
extern int32_t mublasVscal (mublasHandle_t*, const size_t, const cudfcomplex*, cudfcomplex*, const size_t);
extern int32_t mublasDBscal (mublasHandle_t*, const size_t, const double*, double*, const size_t);

// CSRMV
extern int32_t mublasScsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const float*, const float*, const int32_t*, const int32_t*, const float*, const float*, float*);
extern int32_t mublasDcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const double*, const double*, const int32_t*, const int32_t*, const double*, const double*, double*);
extern int32_t mublasCcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const cuFloatComplex*, const cuFloatComplex*, const int32_t*, const int32_t*, const cuFloatComplex*, const cuFloatComplex*, cuFloatComplex*);
extern int32_t mublasZcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const cuDoubleComplex*, const cuDoubleComplex*, const int32_t*, const int32_t*, const cuDoubleComplex*, const cuDoubleComplex*, cuDoubleComplex*);
extern int32_t mublasWcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const cuddreal*, const cuddreal*, const int32_t*, const int32_t*, const cuddreal*, const cuddreal*, cuddreal*);
extern int32_t mublasUcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const cuddcomplex*, const cuddcomplex*, const int32_t*, const int32_t*, const cuddcomplex*, const cuddcomplex*, cuddcomplex*);
extern int32_t mublasBcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const cudfreal*, const cudfreal*, const int32_t*, const int32_t*, const cudfreal*, const cudfreal*, cudfreal*);
extern int32_t mublasVcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const cudfcomplex*, const cudfcomplex*, const int32_t*, const int32_t*, const cudfcomplex*, const cudfcomplex*, cudfcomplex*);
extern int32_t mublasDBcsrmv (mublasHandle_t*, const char, const size_t, const size_t, const size_t, const double*, const double*, const int32_t*, const int32_t*, const double*, const double*, double*);

// cublas compatible (DB)
extern int32_t mublasDBgemmCC (const char, const char, const size_t, const size_t, const size_t, const double*, const double*, const size_t, const double*, const size_t, const double*, double*, const size_t);

#endif
