#ifndef CUDF_H
#define CUDF_H

#include "mublas_devel.h"

/* ***********************************************
   Error Free Translations
   *********************************************** */   

#define SPLITTER_FLOAT 4097.0 //2^12+1
#define SPLITTER_DF 536870913.0 //2^29+1

__device__ __forceinline__ void
cuBTwoSum
	(float a, float b, float &s, float &e)
{
	float v;
	s = __fadd_rn (a, b);
	v = __fsub_rn (s, a);
	e = __fadd_rn ( __fsub_rn(a, __fsub_rn (s, v)), __fsub_rn (b, v));
}

__device__ __forceinline__ void
cuBQuickTwoSum
	(float a, float b, float &s, float &e)
{
	s = __fadd_rn (a, b);
	e = __fsub_rn (b, __fsub_rn (s, a));
}

__device__ __forceinline__ void
cuBSplit
	(float a, float &h, float &l) 
{
	float t;
	t = __fmul_rn (SPLITTER_FLOAT, a);
	h = __fsub_rn (t, __fsub_rn (t, a));
	l = __fsub_rn (a, h);
}

__device__ __forceinline__ void
cuBTwoProd
	(float a, float b, float &p, float &e)
{
	float s1, s2, t1, t2;
	p = __fmul_rn (a, b);
	cuBSplit(a, s1, s2);
	cuBSplit(b, t1, t2);
	e = __fadd_rn (( __fadd_rn (__fsub_rn ( __fmul_rn (s1, t1), p),
					 __fadd_rn (__fmul_rn (s1, t2), __fmul_rn (s2, t1)))),
					 __fmul_rn (s2, t2));
}

__device__ __forceinline__ void
cuBTwoProdFMA
	(float a, float b, float &p, float &e)
{
	p = __fmul_rn (a, b);
	e = __fmaf_rn (a, b, -p);
}


/* ***********************************************
   DF-Arithmetic
   *********************************************** */   

// ----- ADD -------------------------------------

// DF = DF + DF
__device__ __forceinline__ cudfreal
cuBaddV1
	(const cudfreal a, const cudfreal b)
{
	cudfreal t, s, c;
	cuBTwoSum (a.x, b.x, t.x, t.y);
	cuBTwoSum (a.y, b.y, s.x, s.y);
	t.y = __fadd_rn(t.y, s.x);
	cuBQuickTwoSum (t.x, t.y, t.x, t.y);
	t.y = __fadd_rn (t.y, s.y);
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = DF + DF
__device__ __forceinline__ cudfreal
cuBaddV2
	(const cudfreal a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoSum (a.x, b.x, t.x, t.y);
	t.y = __fadd_rn (t.y, __fadd_rn (a.y, b.y));
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = F + DF
__device__ __forceinline__ cudfreal
cuBaddV2
	(const float a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoSum (a, b.x, t.x, t.y);
	t.y = __fadd_rn (t.y, b.y);
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = DF + F
__device__ __forceinline__ cudfreal
cuBaddV2
	(const cudfreal a, const float b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoSum (a.x, b, t.x, t.y);
	t.y = __fadd_rn (t.y, a.y);
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = F + F
__device__ __forceinline__ cudfreal
cuBaddV2
	(const float a, const float b)
{
	cudfreal c;
	cuBTwoSum (a, b, c.x, c.y);
	return c;
}

// ----- SUB -------------------------------------

// DF = DF - DF
__device__ __forceinline__ cudfreal
cuBsubV1
	(const cudfreal a, const cudfreal b)
{
	cudfreal t, s, c;
	cuBTwoSum (a.x, -b.x, t.x, t.y);
	cuBTwoSum (a.y, -b.y, s.x, s.y);
	t.y = __fadd_rn(t.y, s.x);
	cuBQuickTwoSum (t.x, t.y, t.x, t.y);
	t.y = __fadd_rn (t.y, s.y);
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = DF - DF
__device__ __forceinline__ cudfreal
cuBsubV2
	(const cudfreal a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoSum (a.x, -b.x, t.x, t.y);
	t.y = __fadd_rn (t.y, __fsub_rn (a.y, b.y));
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// ----- MUL -------------------------------------

// DF = DF * DF
__device__ __forceinline__ cudfreal
cuBmulV1
	(const cudfreal a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoProdFMA (a.x, b.x, t.x, t.y);
	t.y = __fadd_rn (t.y, __fadd_rn (__fmul_rn(a.x, b.y), __fmul_rn (a.y, b.x)));
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = F * DF
__device__ __forceinline__ cudfreal
cuBmulV1
	(const float a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoProdFMA (a, b.x, t.x, t.y);
	t.y = __fadd_rn (t.y, __fmul_rn (a, b.y));
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = DF * F
__device__ __forceinline__ cudfreal
cuBmulV1
	(const cudfreal a, const float b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoProdFMA (a.x, b, t.x, t.y);
	t.y = __fadd_rn (t.y, __fmul_rn (a.y, b));
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = F * F
__device__ __forceinline__ cudfreal
cuBmulV1
	(const float a, const float b)
{
	cudfreal c;
	cuBTwoProdFMA (a, b, c.x, c.y);
	return c;
}

// DF = DF * DF
__device__ __forceinline__ cudfreal
cuBmulV2
	(const cudfreal a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoProdFMA (a.x, b.x, t.x, t.y);
	t.y = __fmaf_rn(a.x, b.y, __fmaf_rn (a.y, b.x, t.y));
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = F * DF
__device__ __forceinline__ cudfreal
cuBmulV2
	(const float a, const cudfreal b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoProdFMA (a, b.x, t.x, t.y);
	t.y = __fmaf_rn (a, b.y, t.y);
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = DF * F
__device__ __forceinline__ cudfreal
cuBmulV2
	(const cudfreal a, const float b)
{
	cudfreal t;
	cudfreal c;
	cuBTwoProdFMA (a.x, b, t.x, t.y);
	t.y = __fmaf_rn (a.y, b, t.y);
	cuBQuickTwoSum (t.x, t.y, c.x, c.y);
	return c;
}

// DF = DF * DF
__device__ __forceinline__ cudfreal
cuBmulV3
	(const cudfreal a, const cudfreal b)
{
	float m1, t, e;
	cudfreal c;
	m1 = __fmul_rn(a.x, b.y);
	t = __fmaf_rn(a.y, b.x, m1);
	c.x = __fmaf_rn(a.x, b.x, t);
	e = __fmaf_rn(a.x, b.x, -c.x);
	c.y = __fadd_rn(e, t);
	return c;
}

// ----- MAD -------------------------------------

// DF = DF * DF + DF
__device__ __forceinline__ cudfreal
cuBmadfast
	(const cudfreal a, const cudfreal b, const cudfreal c)
{
	cudfreal t, s;
	cudfreal d;
// MULV2: S=A*B
	cuBTwoProdFMA (a.x, b.x, t.x, t.y); //2
	t.y = __fmaf_rn(a.x, b.y, __fmaf_rn (a.y, b.x, t.y)); //2
//	cuBQuickTwoSum (t.x, t.y, s.x, s.y); //3
	s.x = t.x;
	s.y = t.y;
// ADDV2: D=S+C
	cuBTwoSum (s.x, c.x, t.x, t.y); //6
	t.y = __fadd_rn (t.y, __fadd_rn (s.y, c.y)); //2
	cuBQuickTwoSum (t.x, t.y, d.x, d.y); //3
	return d;
}

// ----- SPLIT -------------------------------------

// D->DF
__device__ __forceinline__ cudfreal
double_to_cudfreal
	(const double d)
{
	double t, h, l;
	t = __dmul_rn (SPLITTER_DF, d);
	h = __dsub_rn (t, (__fma_rn (SPLITTER_DF, d, -d)));
	l = __dsub_rn (d, h);
	return make_cudfreal ((float)h, (float)l);
}

// DF->D
__device__ __forceinline__ double
cudfreal_to_double
	(const cudfreal df)
{
	return __dadd_rn((double)df.x, (double)df.y);
}

#endif

