/* ***********************************************
 
	GEMM for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemm_nn_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
 
#ifdef PREC_DB
__global__ void mublasDtoDBmatconv (
	const SIZE_T m,
	const SIZE_T n,
	const double * __restrict__ Ag,
	const SIZE_T lda,
	cudfreal *Bg,
	const SIZE_T ldb
) {
	SIZE_T ix, iy;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T iBy = blockIdx.y;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T nTx = blockDim.x;
	const SIZE_T nTy = blockDim.y;
	ix = iBx * nTx + iTx;
	iy = iBy * nTy + iTy;
	if (ix < m && iy < n) 
		Bg[iy * ldb + ix] = double_to_cudfreal (Ag[iy * lda + ix]);
}
#endif

__host__ int32_t
#ifdef MUBLAS_CC
mublasDBgemmCC (
#else
MUBLAS_GEMM (
	mublasHandle_t* mh,	
#endif
	const char transA,
	const char transB,
	const size_t m,
	const size_t n,
	const size_t k,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const size_t lda,
	const TYPE0 *b,
	const size_t ldb,
	const TYPE0 *beta,
	TYPE0 *c,
	const size_t ldc
) {
#ifdef PREC_DB
	void (*kernel) (const SIZE_T, const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0*, const SIZE_T, const TYPE0, const cudfreal*, const cudfreal*, TYPE0*, const SIZE_T);
#else
	void (*kernel) (const SIZE_T, const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0*, const SIZE_T, const TYPE0, TYPE0*, const SIZE_T);
#endif
	size_t kernel_number = 999;
	size_t use_vector = 0;
	dim3 threads, grid;

	// ---------------------------------------------
	// 1.
	// ---------------------------------------------
#ifdef MUBLAS_CC
	mublasHandle_t *mh;
	mh = (mublasHandle_t *)malloc(sizeof(mublasHandle_t));
	mublasCreateSimple (mh);
#endif
	mublasRecipeDefault (mh);
#ifdef PREC_DB
	cudfreal *dev_Ab, *dev_Bb;
	cudaMalloc ((void **) &dev_Ab, sizeof (cudfreal)* lda * k);
	cudaMalloc ((void **) &dev_Bb, sizeof (cudfreal)* ldb * n);
	mh->Bx = ceilingEx (m / 16., 1);
	mh->By = ceilingEx (k / 16., 1);
	threads = dim3 (16, 16);
	grid = dim3 (mh->Bx, mh->By);
	mublasDtoDBmatconv <<< grid, threads >>> (m, k, a, lda, dev_Ab, lda);
	mh->Bx = ceilingEx (k / 16., 1);
	mh->By = ceilingEx (n / 16., 1);
	grid = dim3 (mh->Bx, mh->By);
	mublasDtoDBmatconv <<< grid, threads >>> (k, n, b, ldb, dev_Bb, ldb);
#endif
	
	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if (
		(((long)a%(128/sizeof(TYPE0)))%16 == 0) && (lda%mh->N_T == 0) &&
		(((long)b%(128/sizeof(TYPE0)))%16 == 0) && (ldb%mh->N_T == 0) &&
		(((long)c%(128/sizeof(TYPE0)))%16 == 0) && (ldc%mh->N_T == 0)
		) {
		use_vector = 1;
	} else {
		use_vector = 0;
	}
	mh->m = m;
	mh->n = n;
	mh->k = k;
	if (m%16||n%16||k%16) use_vector = 0;
	if (m<=64||n<=64||k<=64) use_vector = 0;
	if (VEC4 && (m%64||n%64||k%64)) use_vector = 0;
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Kernel Select
	if (transB == 'N' || transB == 'n') {
		if (transA == 'N' || transA == 'n') {
			mh->BlkDim = 2;
			mh->func_cache = cudaFuncCachePreferShared;
			if (mh->N_T == 4 && use_vector) {
				kernel_number = 140;
				mh->TxStatic = 8;
				mh->TyStatic = 8;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb8x8;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb8x8;
				#ifndef PREC_DB
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb8x8_rb8x8_v4_maxwell;
				#endif
			}
			else if (mh->N_T == 2 && use_vector) {
				kernel_number = 120;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				#ifdef PREC_D
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb8x8;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb8x8;
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb16x16_rb8x8_v2_maxwell;
				#else
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb2x2;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb2x2;
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb16x16_rb2x2_v2_maxwell;
				#endif
			} else {
				kernel_number = 110;
				mh->TxStatic = 16;
				mh->TyStatic = 16;
				mh->func_getShmPerBlk = &getShmPerBlk_GEMM_NN_rb1x1;
				mh->func_getBlkPerGrd = &getBlkPerGrd_GEMM_NN_rb1x1;
				kernel = &MUBLAS_GEMM_NN_DEVICE_tb16x16_rb1x1_s1_maxwell;
			}
		}
	}
	// ---------------------------------------------
		
	// ---------------------------------------------
	// 4. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_GEMM)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		cublasOperation_t tranAcublas = CUBLAS_OP_N;
		cublasOperation_t tranBcublas = CUBLAS_OP_N;
		if (transA == 'T' || transA == 't') tranAcublas = CUBLAS_OP_T;
		if (transA == 'C' || transA == 'c') tranAcublas = CUBLAS_OP_C;
		if (transB == 'T' || transB == 't') tranBcublas = CUBLAS_OP_T;
		if (transB == 'C' || transB == 'c') tranBcublas = CUBLAS_OP_C;
		CUBLAS_GEMM (ch, tranAcublas, tranBcublas, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 5. Thread-Block Size Adjust and Kernel Launch
	cudaFuncGetAttributes (&mh->funcatb, kernel);
	cudaFuncSetCacheConfig (kernel, mh->func_cache);
	cudaDeviceSetSharedMemConfig(cudaSharedMemBankSizeEightByte);
	mublas_optimizer (mh);
	threads = dim3 (mh->Tx, mh->Ty);
	grid = dim3 (mh->Bx, mh->By);
#ifdef PREC_DB
	if (mh->autotuning!=1) kernel <<< grid, threads, 0, mh->streamId >>> (m, n, k, *alpha, a, lda, b, ldb, *beta, dev_Ab, dev_Bb, c, ldc);
	cudaFree (dev_Ab);
	cudaFree (dev_Bb);
#else
	if (mh->autotuning!=1) kernel <<< grid, threads, 0, mh->streamId >>> (m, n, k, *alpha, a, lda, b, ldb, *beta, c, ldc);
#endif
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;

#ifdef MUBLAS_CC
	mublasDestroySimple (mh);
	free (mh);
#endif

	return 0;
}

